import logging
from pathlib import Path
from typing import Any, Union

import ruamel.yaml as yaml
from ruamel.yaml.comments import CommentedMap

log = logging.getLogger('yaml')
log.addHandler(logging.NullHandler())

def dump(*a, **kw) -> str:
    kw['Dumper'] = yaml.RoundTripDumper
    kw['default_flow_style'] = False
    return yaml.dump(*a, **kw)

def load(*a, **kw) -> Any:
    kw['Loader'] = yaml.RoundTripLoader
    return yaml.load(*a, **kw)
    
def read_yaml(path:Union[str,Path]) -> Any:
    path = str(path) # for Path objects
    with open(path) as rfp:
        data = load(rfp)
    return data

def write_yaml(data:Any, path:Union[str,Path]) -> None:
    path = str(path) # for Path objects
    with open(path,'w') as wfp:
        dump(CommentedMap(data), wfp)




import argparse
import logging

from .spec import parse_spec
from .build import Builder
from . import inventory

log = logging.getLogger('dowright')
log.addHandler(logging.NullHandler())

def setup_logging(level):
    level = logging.getLevelName(level.upper())
    log_format = '{asctime} {levelname}: {message}'
    datefmt = '%Y-%m-%d_%H:%M:%S'
    logging.basicConfig(level=level,format=log_format,datefmt=datefmt,style='{')

def get_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('spec', 
                        help='YAML specification for droplet management')
    parser.add_argument('-b','--build',action='store_true')
    parser.add_argument('-w','--wait',action='store_true')
    parser.add_argument('-i','--inventory',action='store_true')
    parser.add_argument('-I','--ips',action='store_true')
    parser.add_argument('-d','--domains',action='store_true')
    parser.add_argument('--destroy',action='store_true')
    parser.add_argument('--loglevel', default='info')

    args = parser.parse_args()

    return args

def main():
    args = get_args()

    setup_logging(args.loglevel)

    spec = parse_spec(args.spec)
    builder = Builder(spec)

    if args.build:
        builder.build_droplets()

    if args.wait:
        builder.wait_for_build()

    if args.ips:
        builder.configure_floating_ips()

    if args.domains:
        builder.configure_domains()
        
    if args.inventory:
        inventory.create_inventory(builder)

    if args.destroy:
        builder.destroy_droplets()

if __name__=='__main__':
    main()



import logging
import copy
from pathlib import Path
from collections import OrderedDict

import jinja2

from . import build

log = logging.getLogger('inventory')
log.addHandler(logging.NullHandler())

class DOWrightInventoryError(Exception):
    pass


def find_private_interface(droplet:dict) -> str:
    for interface in droplet.networks['v4']:
        if interface['type'] == 'private':
            return interface['ip_address']
    raise DOWrightInventoryError(
        'The droplet (%s) has no private network interface'.format(
            droplet.name
        )
    )

def get_inventory_groups(builder: build.Builder,
                         inventory_hostnames:dict) -> dict:
    groups = OrderedDict()
    defaults = builder.spec['inventory'].get('defaults',{})

    for name,group_dict in builder.spec['inventory']['groups'].items():
        is_group_of_hosts = 'hosts' in group_dict
        if is_group_of_hosts:
            group_list = group_dict['hosts']
            group = groups[name] = OrderedDict()
        else:
            # Group of groups
            group_list = group_dict['children']
            group = groups[name+':children'] = OrderedDict()

        group['children'] = OrderedDict()
        group['vars'] = group_dict.get('vars',{}).copy()
        for i,group_item in enumerate(group_list):
            new = defaults.copy()

            name = group_item.pop('name')
            new.update(group_item)

            if is_group_of_hosts:
                # If name in inventory_hostnames, then there is a FQDN
                # for this host that should be used
                name = inventory_hostnames.get(builder.droplet_name(name), name)

            group['children'][name] = new

    return groups


def format_group(name, group):
    items = []
    for item_name, item_vars in group.items():
        var_str = ' '.join('{}={}'.format(k,v) for k,v in item_vars.items())
        items.append('{} {}'.format(item_name, var_str))
    group_str = '{name}{items}\n'.format(
        name='[{}]\n'.format(name) if name else '',
        items='\n'.join(items),
    )
    return group_str
    
def format_group_vars(name, group_vars):
    items = ['{}={}'.format(k,v) for k,v in group_vars.items()]
    vars_str = '[{name}:vars]\n{items}\n'.format(
        name=name, items='\n'.join(items),
    )
    return vars_str
    
def create_inventory(builder: build.Builder) -> None:
    inventory_path = Path(builder.spec['inventory']['name'])\
                     .expanduser().resolve()
    log.info('Creating host inventory: %s', inventory_path)
    droplet_configs = builder.droplet_configurations

    if droplet_configs:
        inventory_groups_str = []

        inventory_hosts = OrderedDict()
        inventory_hostnames = {}
        # Individual droplet hosts
        for dconfig in droplet_configs:
            spec, do_droplet, fips, groups, hosts = map(
                dconfig.get, ['spec','dod','fips','groups','hosts']
            )

            # Use the FQDN if a domain has been registered for the
            # droplet, otherwise use the droplet name
            inv_hostname = do_droplet.name
            if hosts:
                inv_hostname = hosts[0]
                # inventory_hostnames will only have keys for droplets
                # with configured FQDNs
                inventory_hostnames[do_droplet.name] = inv_hostname

            ip = fips[0] if fips else do_droplet.ip_address
            user = 'root'
            private_ip = find_private_interface(do_droplet)
            private_hostname = '{}.private'.format(do_droplet.name)

            inventory_hosts[inv_hostname] = OrderedDict([
                ('ansible_host', ip),
                ('ansible_user', user),
                ('private_ipv4_address', private_ip),
                ('private_hostname', private_hostname),
                ('ansible_python_interpreter', '/usr/bin/python3'),
            ])

        inventory_groups_str.append(
            format_group('', inventory_hosts)
        )

        # Droplet groups
        for group_name, droplet_specs in builder.spec['droplets'].items():
            group = OrderedDict()

            for s in droplet_specs:
                c = builder.droplet_kwargs(s)
                name = inventory_hostnames.get(c['name'], c['name'])
                group[name] = {}
                    
            inventory_groups_str.append(
                format_group(group_name, group)
            )

        # Droplet slug groups (i.e. grouping by OS)
        slugs = OrderedDict(sorted(
            [(c['dod'].image['slug'],[]) for c in droplet_configs],
            key=lambda t:t[0],
        ))
        for c in droplet_configs:
            d = c['dod']
            slugs[d.image['slug']].append(
                inventory_hostnames.get(d.name,d.name)
            )
        for slug_name, group_list in slugs.items():
            inventory_groups_str.append(
                format_group(slug_name, {n:{} for n in group_list})
            )

        groups = get_inventory_groups(builder, inventory_hostnames)
        for group_name, group_dict in groups.items():
            inventory_groups_str.append(
                format_group(group_name, group_dict['children'])
            )
            if group_dict['vars']:
                inventory_groups_str.append(
                    format_group_vars(group_name, group_dict['vars'])
                )
                
            
        inventory_path.write_text(
            '\n\n'.join(inventory_groups_str)
        )

        # template = ansible_hosts_conf_template()
        # inventory_path.write_text(
        #     template.render(
        #         all_droplets=all_droplets,
        #         group_configs=group_configs,
        #         groups=get_inventory_groups(builder),
        #         slugs=slugs,
        #         private_ips=private_ips,
        #     )
        # )
    else:
        log.info('.. No droplets to configure.')


ansible_hosts_conf_template = lambda: jinja2.Template('''
{% for droplet in (all_droplets): -%}
{{ droplet.name }} ansible_host={{ droplet.ip_address }} ansible_user=root private_ipv4_address={{ private_ips[droplet.name] }} private_hostname={{ droplet.name }}.private ansible_python_interpreter=/usr/bin/python3
{% endfor %}

{% for dtype in droplet_configs: -%}
{%   if droplet_configs[dtype]|length -%}
[{{ dtype }}]
{%     for droplet in droplet_configs[dtype]: -%}
{{ droplet.name }}
{%     endfor -%}
{%   endif -%}
{% endfor -%}


{% for slug in slugs -%}
[{{ slug }}]
{%   for droplet in slugs[slug] -%}
{{ droplet.name }}
{%   endfor -%}
{% endfor -%}


{% for gname in groups: -%}
{%   if groups[gname]|length -%}
[{{ gname }}]
{%     for group in groups[gname] -%}
{{ group.name }}
{%     endfor %}
{%   endif -%}
{% endfor %}
''')




import logging
from typing import Any, Union, List
import re
import ipaddress
import copy
import time
from collections import OrderedDict

import jinja2
import digitalocean
from digitalocean import Manager, Droplet
from memoized_property import memoized_property
from gems import composite
import paramiko
from paramiko import SSHClient
from tokenmanager import get_tokens

from . import yaml

log = logging.getLogger('build')
log.addHandler(logging.NullHandler())

class DOWrightBuildError(Exception):
    pass

def get_token(dotoken_type:str) -> Union[composite, str]:
    tokens = get_tokens()
    return tokens.digitalocean[dotoken_type]

def find_record(r, domain):
    test = (r['type'],r['name'],r['data'])
    for record in domain.get_records():
        if test == (record.type, record.name, record.data):
            return record

def is_ip_address(address):
    try:
        ipaddress.ip_address(address)
    except ValueError:
        return False
    return True

def droplet_ssh(droplet:Droplet) -> SSHClient:
    ssh = paramiko.SSHClient()
    ssh.load_system_host_keys()
    ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
    ssh.connect(droplet.ip_address, username='root')
    return ssh

def droplet_file_exists(droplet:Droplet, path:str):
    with droplet_ssh(droplet) as ssh:
        _,stdout,_ = ssh.exec_command(
            'cat {} &> /dev/null && echo OK'.format(path)
        )
        exists = bool(stdout.read())
    return exists
    
cc_udata_template = lambda: jinja2.Template('''\
#cloud-config
runcmd: 
{{ commands }}
''')
def get_cloud_config_user_data(commands:list=None) -> str:
    template = cc_udata_template()
    commands = commands or []

    commands.append('touch /.cloud-config-done')

    return template.render(
        commands=yaml.dump(commands),
    )


class Builder:
    def __init__(self, spec:dict, dry_run:bool=False):
        self.spec = spec
        self.dry_run = dry_run

    @memoized_property
    def token(self) -> str:
        return get_token(self.spec['token'])

    @memoized_property
    def manager(self) -> Manager:
        return Manager(token=self.token)

    @property
    def droplets(self) -> list:
        prefix = self.spec['prefix']
        return [d for d in self.manager.get_all_droplets()
                if d.name.startswith(prefix)]

    @property
    def floating_ips(self) -> list:
        return self.manager.get_all_floating_ips()

    @property
    def domains(self) -> list:
        return self.manager.get_all_domains()

    @property
    def droplet_configurations(self) -> list:
        configs = OrderedDict()
        do_dlut = {d.name: d for d in self.droplets}
        domains = self.domains
        floating_ips = self.floating_ips
        for dgroup, droplets in self.spec['droplets'].items():
            for d in droplets:
                kwargs = self.droplet_kwargs(d, [dgroup])
                do_droplet = do_dlut[kwargs['name']]
                fips = [ip.ip for ip in floating_ips
                        if ip.droplet['name']==kwargs['name']]
                allips = set([do_droplet.ip_address]+fips)
                hostnames = []
                for domain in domains:
                    for r in domain.get_records():
                        if r.data in allips:
                            hostnames.append('.'.join([r.name,domain.name]))
                config = {
                    'spec': kwargs,
                    'dod': do_droplet,
                    'fips': fips,
                    'groups': [dgroup],
                    'hosts': hostnames,
                }
                configs[kwargs['name']] = config
        return configs.values()

    def droplet_name(self, name:str) -> str:
        prefix = self.spec['prefix']
        return '{}-{}'.format(prefix, name)

    def prep_domain_records(self, records:list, droplets:list)->list:
        records = records[:]
        for i,rec in enumerate(records):
            rec = rec.copy()
            if not is_ip_address(rec['data']):
                dname = self.droplet_name(rec['data'])
                droplet = [d for d in droplets if d.name == dname][0]
                rec['data'] = droplet.ip_address
            records[i] = rec
        return records

    def prep_droplet(self, droplet:dict) -> dict:
        new = copy.deepcopy(droplet)
        new['name'] = self.droplet_name(droplet['name'])

        config_commands = new.pop('cloud_config_commands',[])
        if config_commands:
            new['user_data'] = get_cloud_config_user_data(config_commands)

        return new

    def droplet_kwargs(self, droplet:dict, tags:List[str]=None) -> dict:
        kwargs = copy.deepcopy(self.spec.get('defaults',{}))
        droplet = self.prep_droplet(droplet)
        log.debug(droplet)
        kwargs.update(droplet)
        kwargs['token'] = self.token
        if tags:
            if 'tags' in kwargs:
                kwargs['tags'].extend(tags)
            else:
                kwargs['tags'] = tags
        return kwargs

    def build_droplets(self) -> None:
        log.info('Building droplets:')
        existing = set(d.name for d in self.droplets)
        prefix = self.spec['prefix']

        for droplet_tag, droplets in self.spec['droplets'].items():
            tags = [droplet_tag, prefix, ]
            kwargs_list = filter(
                lambda kw:kw['name'] not in existing,
                [self.droplet_kwargs(d, tags) for d in droplets]
            )
            # for kwargs in sorted(kwargs_list, key=lambda d:d['name']):
            for kwargs in kwargs_list:
                log.info('  .. building droplet: %s', kwargs['name'])
                droplet = Droplet(**kwargs)
                if not self.dry_run:
                    droplet.create()
                else:
                    log.info('  .. DRY RUN .. did not build.')

        if not self.dry_run:
            droplets = self.droplets
            dlut = {d.name:d for d in droplets}
            new_droplets = [
                dlut[n] for n in set(d.name for d in droplets)-existing
            ]
            log.info('New droplets created:')
            for d in new_droplets:
                log.info('  .. name: %s ip: %s',d.name, d.ip_address)
        else:
            log.info('DRY RUN .. no droplets created.')

    def destroy_droplets(self)->None:
        droplets = self.droplets
        floating_ips = self.floating_ips
        domains = self.domains

        if droplets:
            log.info('Destroying all (%s) droplets...', len(droplets))

            for ip in floating_ips:
                if ip.droplet and ip.droplet['name'] in {d.name for d in droplets}:
                    log.info('  .. unassigning IP (%s) for droplet (%s)',
                             ip.ip, ip.droplet['name'])
                    if not self.dry_run:
                        ip.unassign()
                    else:
                        log.info('  .. DRY RUN .. did not unassign.')

            for domain_name, records in self.spec['domains'].items():
                domain = [d for d in domains if d.name == domain_name][0]
                records = self.prep_domain_records(records, droplets)
                for r in records:
                    record = find_record(r, domain)
                    if record:
                        log.info('  .. removing domain record: %s %s %s',
                                 record.type, record.name, record.data)
                        if not self.dry_run:
                            record.destroy()
                        else:
                            log.info('  .. DRY RUN .. did not remove.')

            log.info('  .. destroying droplets')
            for d in droplets:
                log.info('    .. {}'.format(d.name))
                if not self.dry_run:
                    d.destroy()
                else:
                    log.info('    .. DRY RUN .. did not destroy.')
        else:
            log.info('No droplets to destroy.')

    def configure_floating_ips(self) -> None:
        droplets = self.droplets
        floating_ips = self.floating_ips

        if 'floating_ips' in self.spec:
            log.info('Configuring floating IPs:')
            for ip,name in self.spec['floating_ips'].items():
                fip = [i for i in floating_ips if i.ip == ip]
                if not fip:
                    raise DOWrightBuildError(
                        'No floating IP with ip: {}'.format(ip)
                    )
                fip = fip[0]
                droplet_name = self.droplet_name(name)
                droplets = [d for d in droplets if d.name==droplet_name]
                if not droplets:
                    raise DOWriteBuildError(
                        'No droplet for this spec with'
                        ' name: {} ({})'.format(name, droplet_name)
                    )
                droplet = droplets[0]
                if ((not fip.droplet) or (fip.droplet and
                                          fip.droplet['id'] != droplet.id)):
                    log.info('  assigning %s to %s', fip.ip, droplet.name)
                    if not self.dry_run:
                        fip.assign(droplet.id)
                    else:
                        log.info('  .. DRY RUN .. no assignment.')
        else:
            log.info('No floating IPs to configure.')

    def configure_domains(self) -> None:
        if 'domains' in self.spec:
            log.info('Configuring domains:')

            droplets = self.droplets
            domains = self.domains
            found = False
            for domain_name, records in self.spec['domains'].items():
                domain = [d for d in domains if d.name == domain_name][0]
                records = self.prep_domain_records(records, droplets)
                for r in records:
                    if not find_record(r, domain):
                        found = True
                        kw = {'type':r['type'], 'name':r['name'],
                              'data':r['data']}
                        log.info('  configuring |type: %s|  |name: %s|'
                                 '  |data: %s|',
                                 kw['type'], kw['name'], kw['data'])
                        if not self.dry_run:
                            domain.create_new_domain_record(**kw)
                        else:
                            log.info('  .. DRY RUN .. not configured.')
            if not found:
                log.info('No new domains to be configured.')
        else:
            log.info('No domains to configure.')

    def droplets_still_building(self):
        return [d for d in self.droplets if d.locked]
    def droplets_still_configuring(self):
        return [d for d in self.droplets
                if not droplet_file_exists(d,'/.cloud-config-done')]

    def wait_for_build(self) -> None:
        log.info('Waiting for all droplets to finish building:')

        not_done = self.droplets_still_building()
        while not_done:
            log.info(' .. waiting on %s droplets', len(not_done))
            time.sleep(5)
            not_done = self.droplets_still_building()

        not_configured = self.droplets_still_configuring()
        while not_configured:
            log.info(' .. waiting for %s droplets to cloud-config',
                     len(not_configured))
            time.sleep(5)
            not_configured = self.droplets_still_configuring()

        log.info('Done.')





import logging
from pathlib import Path
from typing import Any, Union, Dict

# from ansible.inventory import expand_hosts
from . import ansible_expand_hosts as expand_hosts

from .yaml import read_yaml

log = logging.getLogger('spec')
log.addHandler(logging.NullHandler())

class DOWrightSpecError(Exception):
    pass

def verify_spec(spec:dict) -> dict:
    is_str = lambda v: type(v) is str
    is_dict = lambda v: hasattr(v, 'keys')
    required_keys = [
        ('token', is_str),
        ('prefix', is_str),
        ('droplets', is_dict),
    ]
    for key, func in required_keys:
        if not key in spec:
            raise DOWrightSpecError('Key: "%s" required in spec'.format(key))
        if not func(spec[key]):
            raise DOWrightSpecError(
                'Key: "%s" is not required type'.format(key)
            )

    return spec

def parse_spec(path:Union[Path,str]) -> dict:
    log.debug('Reading spec path: %s', path)
    spec = verify_spec(read_yaml(path))
    for dtype,configs in spec['droplets'].items():
        new_configs = []
        log.debug('Found %s configs for droplet type %s', len(configs), dtype)
        for config in configs:
            name = config['name']
            if expand_hosts.detect_range(name):
                log.debug('Found Ansible-style hostname range: %s', name)
                for new_name in expand_hosts.expand_hostname_range(name):
                    new_config = copy.deepcopy(config)
                    new_config['name'] = new_name
                    new_configs.append(new_config)
            else:
                new_configs.append(config)
        log.debug('%s total configs created for droplet type %s',
                  len(new_configs), dtype)
        spec['droplets'][dtype] = new_configs
    return spec




from pathlib import Path

# http://stackoverflow.com/a/27868004/1869370
# from distutils.core import setup
from setuptools import setup

here = Path(__file__).resolve().parent

long_description = Path(here,'README.rst').resolve().read_text()

setup(
    name = 'dowright',
    packages = ['dowright'],
    install_requires=[
        'ruamel.yaml',
        'paramiko',
        'python-digitalocean',
        'memoized_property',
        'gems',
        'tokenmanager',
        'jinja2',
        'ansible',
    ],

    version = '0.0.0a4',
    description = ('Simple YAML-based specification for creation and'
                   ' configuration of DigitalOcean droplets'),
    long_description = long_description,
    url = 'https://bitbucket.org/dogwynn/dowright',

    author="David O'Gwynn",
    author_email="dogywnn@acm.org",
    license='MIT',

    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 3',
    ],
    keywords=('digitalocean yaml specification ansible development'
              ' helper utility'),
)
